#include "Compiler.h"
#include "codeGen/Quadruple.h"
#include "codeGen/RegisterAllocator.h"
#include <filesystem>
#include <iostream>
#include <cassert>

extern std::filesystem::path binaryPath;
const std::string Compiler::extension = "s";
void Compiler::externalCommand(std::filesystem::path lat) {
	return;
	auto obj = lat;
	obj.replace_extension("o");
	auto cmd = std::string("as --32 '") + lat.string() + "' -o '" + obj.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
	auto lib = binaryPath.parent_path() / "lib";
	auto runtime = lib / "runtime.o";
	auto result = obj;
	result.replace_extension("");
	cmd = std::string("ld -m elf_i386 -L'") + lib.string() + "' '" + runtime.string() + "' '" + obj.string() + "' -o '" + result.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
}

string Compiler::compile(Visitable *v) {
	buf <<
		".globl _start\n"
	    ".globl exit\n"
		"_start:\n"
		"	pushl %ebp\n"
		"	movl %esp, %ebp\n"
		"	call main\n"
		"	pushl %eax\n"
		"	call exit\n";

	for (const auto& c : scope->classes) {
		c->calculateSize();
	}

	for (const auto& f : scope->functions) {
		compileFunction(f);
	}

	return buf.str();
}

void Compiler::compileFunction(const FunctionInfoPtr& f) {
	string name = mangleFunctionName(f);
	buf << ".globl \"" << name << "\"\n";

	if (!f->block) return;

	// TODO: jeśli funkcja jest klasowa, ustawić self na ... wartosc z ecx
	// c++ thiscall ecx, stack right-to-left (push c, push b, push a)

	buf << "\"" << name << "\":\n";
	QuadrupleGenerator::Result quadEnv = quadGen.compileFunction(f);
	try {
		RegisterAllocator regGen((quadEnv));
		regGen.allocate();
	} catch(const ParseError &e) {
		printFunction(quadEnv);
		throw ParseError(buf.str() + e.what());
	}

	printFunction(quadEnv);

//	for (auto &b : quadEnv.blocks) {
//		for (auto &q : b->quads) {
//			q->generateAsm(*this);
//		}
//	}

	for (auto &t : temps) t->localOffset = 0;
	for (auto &v : localInfos) v->localOffset = 0;
	temps.clear();
	localInfos.clear();
}

void Compiler::printFunction(QuadrupleGenerator::Result quadEnv) {
	int blkNo = 0;
	for (const auto& b : quadEnv.blocks) {
		buf << "----------------------------------------------------------------------------\n";
		buf << "blok " << blkNo << " " << b->getName() << endl;
		/*buf << "in: ";
		for (auto in : b->in) {
			buf << in->getName() << " ";
		}
		buf << "\nouts: ";
		for (auto out : b->out) {
			buf << out->getName() << " ";
		}
		buf << "\n";*/
		if (!b->flow.in.empty()) {
			buf << "in: ";
			for (const auto& v : b->flow.in) {
				buf << v->name << ", ";
			}
			buf << "\n----------\n";
		}
		if (!b->flow.def.empty()) {
			buf << "def: ";
			for (const auto& v : b->flow.def) {
				buf << v->name << ", ";
			}
			buf << "\n----------\n";
		}
		for (const auto& q : b->quads) {
			auto ret = q->toString();
			if (!ret.empty()) {
				auto len = 0;
				for (const auto &c : ret) {
					len++;
					if (c == '\t') len += 8 - (len % 8);
				}
				for (; len < 70; len++) ret += ' ';
				buf << ret << " alive: ";
				for (auto var : q->aliveAfter)
					buf << var->name << " ";
				buf << endl;
				q->generateAsm(*this);
			}
		}
		if (!b->flow.use.empty()) {
			buf << "----------\nuse: ";
			for (const auto &v : b->flow.use) {
				buf << v->name << ", ";
			}
			buf << "\n";
		}
		if (!b->flow.out.empty()) {
			buf << "----------\nout: ";
			for (const auto &v : b->flow.out) {
				buf << v->name << ", ";
			}
			buf << "\n";
		}
		blkNo++;
	}
}

const char* const Compiler::regs[6] = {"eax", "ebx", "ecx", "edx", "esi", "edi"};

string Compiler::getRef(VariablePtr &v) {
	if (v->registerColor == -1) {
		if (!v->localOffset) {
			if (v->info) {
				if (!v->info->localOffset) {
					localInfos.emplace(v->info);
					v->info->localOffset = 4 * (temps.size() + localInfos.size());
				}
				v->localOffset = v->info->localOffset;
			} else {
				temps.emplace(v);
				v->localOffset = 4 * (temps.size() + localInfos.size());
			}
		}
		return to_string(v->localOffset) + "(%ebp)";
	} else {
		return Register(v);
	}
}

void Compiler::generateQJumpCond(QJumpCond &q) {
//	return QJump::generateAsm();

}

void Compiler::generateQAlloc(QAlloc &q) {


}

void Compiler::generateQWrite(QWrite &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQAccess(QAccess &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQReturn(QReturn &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQCall(QCall &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQParam(QParam &q) {
//	return Quadruple::generateAsm();
	auto reg = Register(q.param);
	if (q.num < 0) {
		auto offset = 4 * q.num;
		if (reg == 0) {
			q.param->localOffset = offset;
		} else {
			append("MOVL", to_string(offset)+"(%ebp)");
		}
	} else {
		append("PUSHL", getRef(q.param));
	}
}

void Compiler::generateQJump(QJump &q) {
//	return Quadruple::generateAsm();
//	append("JMP", q.block.la)

}

void Compiler::generateQAssign(QAssign &q) {
	Register tg(q.loc);
	if (q.args.size() == 1) {
		auto &arg = q.args[0];
		switch (q.op.op) {
			case Op::Not: {
				auto reg = Register(arg);
				if (reg == 0) {
					// avoids extra MOV when arg is already in register
					moveTo(arg, tg);
					reg = tg;
				}
				append("TEST", reg, reg);
				append("SETZ", tg);
				break;
			}
			case Op::Copy: {
				if (arg->constExpr) {
					// TODO
				} else {
					moveTo(arg, tg);
				}
				break;
			}
			case Op::Neg: {
				moveTo(arg, tg);
				append("NEG", tg);
				break;
			}
			default: throw runtime_error("Unexpected operand");
		}
	} else {
		auto &i = q.args[0], &j = q.args[1];
		if (q.op == Op::Div || q.op == Op::Mod) {
			auto edx = Register("edx");
			bool backupEdx = edx != tg && edx != Register(i);
			if (backupEdx) append("MOVL", edx, "4(%esp)");
			auto jLoc = edx != Register(j) ? getRef(j) : "4(%esp)";
			moveTo(i, 0);
			append("CLTD"); // edx lost
			append("IDIVL", jLoc);
			moveTo(q.op == Op::Mod ? edx : 0, getRef(q.loc));
			// restore unnecessary if j was in edx and becomes dead and was NOT coalesced
			if (backupEdx) append("MOVL", "4(%esp)", edx);
			return;
		} else if (q.op < Op::CMP) {
			moveTo(i, tg);
			append(q.op, getRef(i), tg);
		} else {
			auto loc = getRef(i);
			if (Register(i) == 0 && Register(j) == 0) {
				moveTo(i, 0);
				loc = "%eax";
			}

			append("CMPL", loc, getRef(j));
			append(q.op, tg);
		}
	}
	if (tg == 0) {
		append("MOVL", tg, getRef(q.loc));
	}
}

void Compiler::generateQLabel(QLabel &q) {
//	return Quadruple::generateAsm();

}
